#include "math_util.h"

#include <cstdint>
#include <cmath>
#include <limits>
#include "iostream"

namespace Localization
{
namespace PointLoc
{

float MathUtil::InvSqrt(float x) noexcept(false) {
  if (fabs(x) < std::numeric_limits<float>::epsilon()) {
//    throw InvalidArgumentException("Divided by zero");
    std::cerr<<"InvSqrt():Divided by zero,return 1"<<std::endl;
    return 1;
  }
  float xhalf{0.5F * x};
  // get bits for floating value
  std::int32_t i{*(static_cast<std::int32_t*>(static_cast<void*>(&x)))};
  // gives initial guess
  i = 0x5f37642f - (i >> 1);
  // convert bits back to float
  x = *(static_cast<float*>(static_cast<void*>(&i)));
  // Newton step
  x = x * (1.5F - xhalf * x * x);
  return x;
}

float MathUtil::InvSqrtPrecise(float x) noexcept(false) {
  if (fabs(x) < std::numeric_limits<float>::epsilon()) {
//    throw InvalidArgumentException("Divided by zero");
    std::cerr<<"InvSqrtPrecise():Divided by zero,return 1"<<std::endl;
    return 1;
  }
  float xhalf{0.5F * x};
  // get bits for floating value
  std::int32_t i{*(static_cast<std::int32_t*>(static_cast<void*>(&x)))};
  // gives initial guess
  i = 0x5f37642f - (i >> 1);
  // convert bits back to float
  x = *(static_cast<float*>(static_cast<void*>(&i)));
  // Newton step
  x = x * (1.5F - xhalf * x * x);
  x = x * (1.5F - xhalf * x * x);
  return x;
}

double MathUtil::InvSqrt(double x) noexcept(false) {
  if (fabs(x) < std::numeric_limits<double>::epsilon()) {
//    throw InvalidArgumentException("Divided by zero");
    std::cerr<<"InvSqrt():Divided by zero,return 1"<<std::endl;
    return 1;
  }
  double xhalf{0.5 * x};
  // get bits for floating value
  std::int64_t i{*(static_cast<std::int64_t*>(static_cast<void*>(&x)))};
  // gives initial guess
  i = 0x5fe6ec85e7de30da - (i >> 1);
  // convert bits back to double
  x = *(static_cast<double*>(static_cast<void*>(&i)));
  // Newton step
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  return x;
}

double MathUtil::InvSqrtPrecise(double x) noexcept(false) {
  if (fabs(x) < std::numeric_limits<double>::epsilon()) {
//    throw InvalidArgumentException("Divided by zero");
    std::cerr<<"InvSqrtPrecise():Divided by zero,return 1"<<std::endl;
    return 1;
  }
  double xhalf{0.5 * x};
  // get bits for floating value
  std::int64_t i{*(static_cast<std::int64_t*>(static_cast<void*>(&x)))};
  // gives initial guess
  i = 0x5fe6ec85e7de30da - (i >> 1);
  // convert bits back to double
  x = *(static_cast<double*>(static_cast<void*>(&i)));
  // Newton step
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  x = x * (1.5 - xhalf * x * x);
  return x;
}

} // namespace PointLoc
} // namespace Localization
/* vim: set ts=2 sw=2 sts=2 tw=100 et: */
